#include <TRandom.h>

// -----------------------------------------------------------------------------
/// \brief Throw a random direction on the sphere, weighted by an exposure
///        stored in a skymap object.
// -----------------------------------------------------------------------------
template < class T > void
Event::GenerateEventFromSkymap(const Healpix_Map< T > &map, pointing &dir)
{
  int pixel;

  do 
  {
    // Throw a random direction on the sphere.
    do {
      dir.phi   = 360. * gRandom->Rndm() * degr2rad;
      dir.theta = 180. * gRandom->Rndm() * degr2rad;
    }
    while (gRandom->Rndm() > sin(dir.theta));

    // Accept the direction, weighted by the skymap.
    pixel = map.ang2pix(dir);
  } 
  while (gRandom->Rndm() > map[pixel]);
}


// -----------------------------------------------------------------------------
/// \brief Throw a random pixel (direction) on the sphere, weighted by an 
///        exposure stored in a skymap object.
// -----------------------------------------------------------------------------
template < class T > int
Event::GeneratePixelFromSkymap(const Healpix_Map< T > &map)
{
  int pixel;
  const int N = map.Npix();

  // Pick a random direction on the sphere, weighted by the skymap.
  do {
    pixel = gRandom->Integer(N);
  }
  while (gRandom->Rndm() > map[pixel]);

  return pixel;
}


// -----------------------------------------------------------------------------
/// \brief Generate N random points on the sphere, weighted by an exposure 
///        stored in a skymap object.
// -----------------------------------------------------------------------------
template < class T > void
Event::GenerateEventListFromSkymap(
  const Healpix_Map< T > &map, const size_t N, PointArray &parray)
{
  pointing dir;
  parray.clear();

  for (size_t i = 0; i < N; ++i) {
    GenerateEventFromSkymap(map, dir);
    parray.push_back(dir);
  }
}


// -----------------------------------------------------------------------------
/// \brief Generate N random pixels (directions) on the sphere, weighted by an
///        exposure stored in a skymap object.
// -----------------------------------------------------------------------------
template < class T > void
Event::GeneratePixelListFromSkymap(
  const Healpix_Map< T > &map, const size_t N, PixelArray &parray)
{
  parray.clear();

  for (size_t i = 0; i < N; ++i) {
    parray.push_back(GeneratePixelFromSkymap(map));
  }
}


// -----------------------------------------------------------------------------
/// \brief Given a skymap and input pixel, generate another pixel a set
///        angular distance from the first.
// -----------------------------------------------------------------------------
template < class T > int
Event::GeneratePixelNearPosition(
  const Healpix_Map< T > &map, const int pixel, const double angdist)
{
  pointing dir;
  GenerateEventNearPosition(map.pix2ang(pixel), angdist, dir);
  return map.ang2pix(dir);
}


// -----------------------------------------------------------------------------
/// \brief Given a skymap and input pixel, generate another pixel within a disc
///        of radius angdist centered on the first pixel.
// -----------------------------------------------------------------------------
template < class T > int
Event::GeneratePixelWithinDisc(
  const Healpix_Map< T > &map, const int pixel, const double angdist)
{
  pointing dir;
  GenerateEventWithinDisc(map.pix2ang(pixel), angdist, dir);
  return map.ang2pix(dir);
}


// -----------------------------------------------------------------------------
/// \brief Convert a list of map pixels to a list of pointing directions.
// -----------------------------------------------------------------------------
template < class T> void
Event::PixelToPointArray(
  const Healpix_Map< T > &map, const PixelArray &pxarray, PointArray &ptarray)
{
  ptarray.clear();
  PixelArrayCIter pixIt;
  for (pixIt = pxarray.begin(); pixIt != pxarray.end(); ++pixIt)
    ptarray.push_back(map.pix2ang(*pixIt));
}


// -----------------------------------------------------------------------------
/// \brief Convert a list of pointing directions to a list of map pixels.
// -----------------------------------------------------------------------------
template < class T> void
Event::PointToPixelArray(
  const Healpix_Map< T > &map, const PointArray &ptarray, PixelArray &pxarray)
{
  pxarray.clear();
  PointArrayCIter pointIt;
  for (pointIt = ptarray.begin(); pointIt != ptarray.end(); ++pointIt)
    pxarray.push_back(map.ang2pix(*pointIt));
}

