
#include <healpix-cxx/cxxsupport/pointing.h>

#include "Geometry.h"
using namespace Geo;


// -----------------------------------------------------------------------------
/// \brief Find the angular separation between two unit vectors.
// -----------------------------------------------------------------------------
double
Geo::AngularDistance(const vec3 &v1, const vec3 &v2)
{
   vec3 r1 = v1;   r1.Normalize();
   vec3 r2 = v2;   r2.Normalize();
   double dist = 0.0;
   double sprod = dotprod(r1, r2);

   if (sprod > 0.999) {             // Vectors almost parallel.
      vec3 vdiff  = r1 - r2;
      double diff = sqrt(dotprod(vdiff, vdiff));
      dist = 2.0 * asin(0.5 * diff);
   }
   else if (sprod < -0.999) {       // Vectors almost antiparallel.
      vec3 vsum  = r1 + r2;
      double sum = sqrt(dotprod(vsum, vsum));
      dist = pi - 2.0 * asin(0.5 * sum);
   }
   else {
      dist = acos(sprod);
   }

   return dist;
}


// -----------------------------------------------------------------------------
/// \brief Find the angular distance between two points on the unit sphere.
// -----------------------------------------------------------------------------
double
Geo::AngularDistance(const pointing &p1, const pointing &p2)
{
  return Geo::AngularDistance(p1.to_vec3(), p2.to_vec3());
}


// -----------------------------------------------------------------------------
/// \brief Find the angular distance between two points on the sphere.
// -----------------------------------------------------------------------------
double
Geo::AngularDistance(
  const double th1, const double ph1, const double th2, const double ph2)
{
  return Geo::AngularDistance(pointing(th1, ph1), pointing(th2, ph2));
}


// -----------------------------------------------------------------------------
/// \brief Rotate a vector out of a coordinate system whose z-axis is given by
///        the pointing direction zenith.
// -----------------------------------------------------------------------------
void
Geo::Rotate(const pointing &zenith, vec3 &v)
{
  // Rotate a vector OUT of the coordinates with a given zenith.
  double costh = cos(zenith.theta);
  double cosph = cos(zenith.phi);
  double sinth = sin(zenith.theta);
  double sinph = sin(zenith.phi);

  // Rotation: -theta about y, followed by -phi about z.
  v = vec3(
     costh * cosph * v.x  -  sinph * v.y  + sinth * cosph * v.z,
     costh * sinph * v.x  +  cosph * v.y  + sinth * sinph * v.z,
    -sinth * v.x  +  costh * v.z
  );
}

