
#include <iomanip>
#include <iostream>
using namespace std;

// HEALPix
#include <healpix-cxx/cxxsupport/constants.h>
#include <healpix-cxx/cxxsupport/fitshandle.h>
#include <healpix-cxx/healpix/healpix_map.h>
#include <healpix-cxx/healpix/healpix_map_fitsio.h>
#include <healpix-cxx/cxxsupport/pointing.h>

// ROOT
#include <RVersion.h>
#include <TApplication.h>
#include <TAxis.h>
#include <TCanvas.h>
#include <TColor.h>
#include <TGraph.h>
#include <TH1.h>
#include <TH2.h>
#include <TMath.h>
#include <TMultiGraph.h>
#include <TPaveText.h>
#include <TStyle.h>

#include "skymap_functions.h"


// -----------------------------------------------------------------------------
/// \brief Style functions.
// -----------------------------------------------------------------------------
void
SetPrettyStyle()
{
  // Canvas printing details: white bg, no borders.
  gStyle->SetCanvasColor(0);
  gStyle->SetCanvasBorderMode(0);
  gStyle->SetCanvasBorderSize(0);

  // Canvas frame printing details: white bg, no borders.
  gStyle->SetFrameFillColor(0);
  gStyle->SetFrameBorderMode(0);
  gStyle->SetFrameBorderSize(0);

  // Plot title details: centered, no bg, no border, nice font.
  gStyle->SetTitleX(0.1);
  gStyle->SetTitleW(0.8);
  gStyle->SetTitleBorderSize(0);
  gStyle->SetTitleFillColor(0);

  // Font details for titles and labels.
  gStyle->SetTitleFont(42, "xyz");
  gStyle->SetTitleFont(42, "pad");
  gStyle->SetLabelFont(42, "xyz");
  gStyle->SetLabelFont(42, "pad");

  // Details for stat box.
  gStyle->SetStatColor(0);
  gStyle->SetStatFont(42);
  gStyle->SetStatBorderSize(1);
  gStyle->SetStatX(0.975);
  gStyle->SetStatY(0.9);

  // gStyle->SetOptStat(0);
}

void
CreateColorTable(UInt_t Number, Double_t* Length, 
  Double_t* Red, Double_t* Green, Double_t* Blue, UInt_t NColors)
{
  #if ROOT_VERSION_CODE >= ROOT_VERSION(5,16,0)
    TColor::CreateGradientColorTable(Number, Length, Red, Green, Blue, NColors);
  #else
    gStyle->CreateGradientColorTable(Number, Length, Red, Green, Blue, NColors);
  #endif
}

void
SetRootPalette(const Int_t pal = 0)
{
  const Int_t NRGBs = 5;
  const Int_t NCont = 255;

  if (pal == 0) {
    cout << "\n  ROOT Palette Function\n  Usage: SetRootPalette(pal)\n\n"
         << "    o pal =  0: print help\n"
         << "    o pal =  1: rainbow\n"
         << "    o pal =  2: reverse-rainbow\n"
         << "    o pal =  3: amber\n"
         << "    o pal =  4: reverse-amber\n"
         << "    o pal =  5: blue/white\n"
         << "    o pal =  6: white/blue\n"
         << "    o pal =  7: red temperature\n"
         << "    o pal =  8: reverse-red temperature\n"
         << "    o pal =  9: green/white\n"
         << "    o pal = 10: white/green\n"
         << "    o pal = 11: orange/blue\n"
         << "    o pal = 12: blue/orange\n"
         << "    o pal = 13: white/black\n"
         << "    o pal = 14: black/white\n"
         << endl;
  }
  else if (pal == 1) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 0.00, 0.00, 0.87, 1.00, 0.51 };
    Double_t green[NRGBs] = { 0.00, 0.81, 1.00, 0.20, 0.00 };
    Double_t blue[NRGBs]  = { 0.51, 1.00, 0.12, 0.00, 0.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 2) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 0.51, 1.00, 0.87, 0.00, 0.00 };
    Double_t green[NRGBs] = { 0.00, 0.20, 1.00, 0.81, 0.00 };
    Double_t blue[NRGBs]  = { 0.00, 0.00, 0.12, 1.00, 0.51 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 3) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 0.17, 0.39, 0.62, 0.79, 1.00 };
    Double_t green[NRGBs] = { 0.01, 0.02, 0.39, 0.68, 1.00 };
    Double_t blue[NRGBs]  = { 0.00, 0.09, 0.18, 0.09, 0.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 4) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 1.00, 0.79, 0.62, 0.39, 0.17 };
    Double_t green[NRGBs] = { 1.00, 0.68, 0.39, 0.02, 0.01 };
    Double_t blue[NRGBs]  = { 0.00, 0.09, 0.18, 0.09, 0.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 5) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 0.00, 0.00, 0.00, 0.38, 1.00 };
    Double_t green[NRGBs] = { 0.00, 0.00, 0.38, 0.76, 1.00 };
    Double_t blue[NRGBs]  = { 0.00, 0.47, 0.83, 1.00, 1.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 6) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 1.00, 0.38, 0.00, 0.00, 0.00 };
    Double_t green[NRGBs] = { 1.00, 0.76, 0.38, 0.00, 0.00 };
    Double_t blue[NRGBs]  = { 1.00, 1.00, 0.83, 0.47, 0.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 7) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 0.00, 0.50, 0.89, 0.95, 1.00 };
    Double_t green[NRGBs] = { 0.00, 0.00, 0.27, 0.71, 1.00 };
    Double_t blue[NRGBs]  = { 0.00, 0.00, 0.00, 0.40, 1.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 8) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 1.00, 0.95, 0.89, 0.50, 0.00 };
    Double_t green[NRGBs] = { 1.00, 0.71, 0.27, 0.00, 0.00 };
    Double_t blue[NRGBs]  = { 1.00, 0.40, 0.00, 0.00, 0.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 9) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 0.00, 0.00, 0.38, 0.75, 1.00 };
    Double_t green[NRGBs] = { 0.00, 0.35, 0.62, 0.85, 1.00 };
    Double_t blue[NRGBs]  = { 0.00, 0.00, 0.00, 0.47, 1.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 10) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 1.00, 0.75, 0.38, 0.00, 0.00 };
    Double_t green[NRGBs] = { 1.00, 0.85, 0.62, 0.35, 0.00 };
    Double_t blue[NRGBs]  = { 1.00, 0.47, 0.00, 0.00, 0.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 11) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 0.75, 1.00, 0.24, 0.00, 0.00 };
    Double_t green[NRGBs] = { 0.24, 1.00, 0.75, 0.18, 0.00 };
    Double_t blue[NRGBs]  = { 0.00, 0.62, 1.00, 0.68, 0.12 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 12) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 0.00, 0.00, 0.24, 1.00, 0.75 };
    Double_t green[NRGBs] = { 0.00, 0.18, 0.75, 1.00, 0.24 };
    Double_t blue[NRGBs]  = { 0.12, 0.68, 1.00, 0.62, 0.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 13) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t green[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t blue[NRGBs]  = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
  else if (pal == 14) {
    Double_t stops[NRGBs] = { 0.00, 0.34, 0.61, 0.84, 1.00 };
    Double_t red[NRGBs]   = { 1.00, 0.84, 0.61, 0.34, 0.00 }; 
    Double_t green[NRGBs] = { 1.00, 0.84, 0.61, 0.34, 0.00 };
    Double_t blue[NRGBs]  = { 1.00, 0.84, 0.61, 0.34, 0.00 };
    CreateColorTable(NRGBs, stops, red, green, blue, NCont);
    gStyle->SetNumberContours(NCont);
  }
}

// Functions associated with map projections...
// -----------------------------------------------------------------------------
/// \brief Transcendental equation for intermediate Mollweide parameter gamma.
// -----------------------------------------------------------------------------
Double_t
f(Double_t g, Double_t th)
{
   return (2.0 * g + sin(2.0 * g) - TMath::Pi() * sin(th));
}

// -----------------------------------------------------------------------------
/// \brief First derivative of eq. for intermediate Mollweide parameter gamma.
// -----------------------------------------------------------------------------
Double_t
df(Double_t g, Double_t th)
{
   return (2.0 + 2.0 * cos(2.0 * g));
}

// -----------------------------------------------------------------------------
/// \brief Newton-Raphson root finding algorithm used to solve for gamma.
// -----------------------------------------------------------------------------
Double_t
root_find(
   Double_t (*f)(Double_t, Double_t),
   Double_t (*df)(Double_t, Double_t),
   Double_t a,
   Double_t b,
   Double_t th
   )
{
   static Double_t precision = 1.0e-4;
   Double_t dx = b - a;
   Double_t x0 = (a + b) / 2.0;
   Double_t x1;

   do {
      x1 = x0 - f(x0, th) / df(x0, th);
      dx = x1 - x0;
      x0 = x1;
   } while (fabs(dx) > precision);

   return x0;
}

// -----------------------------------------------------------------------------
/// \brief Mollweide projection: long,lat --> x,y
// -----------------------------------------------------------------------------
void
project(const Double_t l, const Double_t b, Double_t *x, Double_t *y)
{
  Double_t g = root_find(f, df, -0.5*TMath::Pi(), 0.5*TMath::Pi(), b);
  *x = 2. * TMath::Sqrt(2.) * l * TMath::Cos(g) / TMath::Pi();
  *y = TMath::Sqrt(2.) * TMath::Sin(g);
}

// -----------------------------------------------------------------------------
/// \brief Inverse Mollweide projection: x,y --> long,lat
// -----------------------------------------------------------------------------
void
inverse_project(const Double_t x, const Double_t y, Double_t *l, Double_t *b)
{
  Double_t g = TMath::ASin(y / TMath::Sqrt(2.));
  *l = TMath::Pi() * x / (2.*TMath::Sqrt(2.) * TMath::Cos(g));
  *b = TMath::ASin((2*g + TMath::Sin(2*g)) / TMath::Pi());
}

// -----------------------------------------------------------------------------
/// \brief Fill and draw function.
// -----------------------------------------------------------------------------
void
exposure(const Char_t *bgfile, const Int_t nLBins=1800, const Int_t nBBins=900)
{
  Healpix_Map< double > bgmap;
  read_Healpix_map_from_fits(bgfile, bgmap);
  rescale_skymap_to_one(bgmap);
  cout << "Read in skymap with " << bgmap.Npix() << " pixels.\n\n";

  const Double_t deg = TMath::DegToRad();

  TH2D *h2 = new TH2D("h2", "", nLBins, -2.83, 2.83, nBBins, -1.42, 1.42);
  TAxis *xaxis = h2->GetXaxis();
  TAxis *yaxis = h2->GetYaxis();

  Int_t line, count, pixel;
  pointing direction;
  Double_t x, y, l, b, acc;

  for (Int_t i = 1; i <= h2->GetNbinsX(); ++i) {
    for (Int_t j = 1; j <= h2->GetNbinsY(); ++j) {
      x = xaxis->GetBinCenter(i);
      y = yaxis->GetBinCenter(j);

      inverse_project(x, y, &l, &b);
  
      if (TMath::Abs(l) <= 180.*deg && TMath::Abs(b) <= 90.*deg) {
        
        // Convert latitude to zenith.
        // Make sure the plot runs from:  +180  +90   0  -90  -180.
        direction = pointing(90.*deg - b, -l);
        pixel = bgmap.ang2pix(direction);
        acc = bgmap[pixel];
        if (acc > 0.) {
          h2->Fill(x, y, acc);
        }
      }
    }
  }

  // Grid lines.
  const Int_t nlon = 9;
  const Int_t nlat = 7;
  const Double_t dlon = 360*deg / (nlon - 1);
  const Double_t dlat = 180*deg / (nlat - 1);

  TGraph *grlon   = new TGraph[nlon];
  TGraph *grlat   = new TGraph[nlat];
  TMultiGraph *mg = new TMultiGraph;

  // Lines of longitude.
  line = 0;
  for (l = -180*deg; l <= 180*deg; l += dlon) {
    count = 0;
    for (Double_t sinb = -1.; sinb <= 1.; sinb += 0.1) {
      b = TMath::ASin(sinb);
      project(l, b, &x, &y);
      grlon[line].SetPoint(count++, x, y);
    }
    grlon[line].SetLineColor(13);
    mg->Add(&grlon[line], "c");
    ++line;
  }

  // Lines of latitude.
  line = 0;
  for (b = -90*deg; b <= 90*deg; b += dlat) {
    count = 0;
    for (l = -180*deg; l <= 180*deg; l += 12*deg) {
      project(l, b, &x, &y);
      grlat[line].SetPoint(count++, x, y);
    }
    grlat[line].SetLineColor(13);
    mg->Add(&grlat[line], "c");
    ++line;
  }

  // Draw everything with a nice aspect ratio.
  TCanvas *c = new TCanvas("c", "Exposure Map", 10, 10, 960, 500);
  c->SetFrameLineColor(0);
  c->SetFrameBorderSize(0);
  c->SetLeftMargin(0.02);
  c->SetTopMargin(0.06);
  c->SetBottomMargin(0.06);
  c->SetRightMargin(0.12);

  gStyle->SetOptStat(0);
  TH1D *hframe = new TH1D("", "", 10, -2.83, 2.83);
  hframe->SetMinimum(-1.42);
  hframe->SetMaximum(+1.42);
  TAxis *axis = hframe->GetXaxis();
  axis->SetAxisColor(0);
  axis->SetLabelColor(0);
  axis = hframe->GetYaxis();
  axis->SetAxisColor(0);
  axis->SetLabelColor(0);
  hframe->Draw();

  axis = h2->GetZaxis();
  axis->CenterTitle();
  axis->SetTitle("relative exposure");
  axis->SetTitleOffset(0.9);
  h2->Draw("colzsame");
  mg->Draw();

  TPaveText *pt = new TPaveText(-2.4, -1.3, -2.3, -1.2);
  pt->SetLineColor(0);
  pt->SetFillColor(0);
  pt->SetTextSize(0.05);
  pt->SetTextFont(42);
  pt->AddText("+180#circ");
  pt->Draw();

  pt = new TPaveText(2.3, -1.3, 2.4, -1.2);
  pt->SetLineColor(0);
  pt->SetFillColor(0);
  pt->SetTextSize(0.05);
  pt->SetTextFont(42);
  pt->AddText("-180#circ");
  pt->Draw();
}


int main(int argc, char *argv[])
{
  if (argc !=2 ) {
    cerr << "\n\tExposure Map Plotter:\n\n"
         << "Usage: " << argv[0] << " bgfile\n\n";

    return 1;
  }

  TApplication app("App", &argc, argv);

  SetPrettyStyle();
  SetRootPalette(1);

  exposure(app.Argv(1));

  app.Run();

  return 0;
}

