
// -----------------------------------------------------------------------------
/// \brief Normalize a skymap such that its integral is unity.
// -----------------------------------------------------------------------------
template <class T> void
normalize_skymap(Healpix_Map<T> &map)
{
   // Integrate (sum) over the skymap;
   // use the sum as a normalization factor.
   int i, numPix = map.Npix();
   T norm = (T)0;

   for (i = 0; i < numPix; ++i) norm   += map[i];
   for (i = 0; i < numPix; ++i) map[i] /= norm;

   return;
}

// -----------------------------------------------------------------------------
/// \brief Rescale a skymap by a user-defined amount.
// -----------------------------------------------------------------------------
template <class T> void
rescale_skymap(Healpix_Map<T> &map, const T &scale)
{
   int i, numPix = map.Npix();
   for (i = 0; i < numPix; ++i) map[i] *= scale;
}

// -----------------------------------------------------------------------------
/// \brief Rescale a skymap such that its values run from zero to one.
// -----------------------------------------------------------------------------
template <class T> void
rescale_skymap_to_one(Healpix_Map<T> &map)
{
   double min, max;
   map.minmax(min, max);
   rescale_skymap(map, 1.0 / max);
}

// -----------------------------------------------------------------------------
/// \brief Multiply two skymaps and output their product.
// -----------------------------------------------------------------------------
template <class T> Healpix_Map<T>
product_of_skymaps(const Healpix_Map<T> &m1, const Healpix_Map<T> &m2)
{
   if (m1.Npix() != m2.Npix()) {
      std::cerr << " * ERROR: product_of_skymaps()\n"
                << "   Skymaps have different pixelization.\n"
                << "   Returning skymap 1...\n";
      return m1;
   }
   else if (m1.Scheme() != m2.Scheme()) {
      std::cerr << " * ERROR: product_of_skymaps()\n"
                << "   Skymaps have a different ordering scheme.\n"
                << "   Returning skymap 1...\n";
      return m1;
   }

   int numPix  = m1.Npix();

   Healpix_Map<T> mprod;
   mprod.SetNside(m1.Nside(), m1.Scheme());
   for (int i = 0; i < numPix; ++i) mprod[i] = m1[i] * m2[i];

   return mprod;
}

// -----------------------------------------------------------------------------
/// \brief Add two skymaps and output their sum.
// -----------------------------------------------------------------------------
template <class T> Healpix_Map<T>
sum_of_skymaps(const Healpix_Map<T> &m1, const Healpix_Map<T> &m2)
{
   if (m1.Npix() != m2.Npix()) {
      std::cerr << " * ERROR: product_of_skymaps()\n"
                << "   Skymaps have different pixelization.\n"
                << "   Returning skymap 1...\n";
      return m1;
   }
   else if (m1.Scheme() != m2.Scheme()) {
      std::cerr << " * ERROR: product_of_skymaps()\n"
                << "   Skymaps have a different ordering scheme.\n"
                << "   Returning skymap 1...\n";
      return m1;
   }

   int numPix  = m1.Npix();

   Healpix_Map<T> msum;
   msum.SetNside(m1.Nside(), m1.Scheme());
   for (int i = 0; i < numPix; ++i) msum = m1[i] + m2[i];

   return msum;
}

